#!/usr/bin/env python3
# License: BSD-3-Clause

import json
import logging
import re

import numpy as np
import requests
from astropy.time import Time

BASE_GCN_URL = "https://gcn.gsfc.nasa.gov/gcn3"


def gcn_url(gcn_number):
    """ """
    return f"{BASE_GCN_URL}/{gcn_number}.gcn3"


class ParsingError(Exception):
    """Base class for parsing error"""

    pass


def find_gcn_no(base_nu_name: str):
    """
    Trick the webpage into giving us results
    """
    endpoint = (
        "https://heasarc.gsfc.nasa.gov/wsgi-scripts/tach/gcn_v2/tach.wsgi/graphql_fast"
    )

    # hard code missing entries
    if base_nu_name == "IC220405B":
        return 31839

    querystr = (
        '{ allEventCard( name: "'
        + base_nu_name
        + '" ) {edges {node { id_ event } } } }'
    )
    r = requests.post(
        endpoint,
        data={
            "query": querystr,
            "Content-Type": "application/json",
        },
    )
    res = json.loads(r.text)

    if res["data"]["allEventCard"]["edges"]:
        event_id = res["data"]["allEventCard"]["edges"][0]["node"]["id_"]

        querystr = (
            "{ allCirculars ( evtid:"
            + event_id
            + " ) { totalCount edges { node { id id_ received subject "
            "evtidCircular{ event } cid evtid oidCircular{ telescope detector "
            "oidEvent{ wavelength messenger } } } } } }"
        )

        r = requests.post(
            endpoint,
            data={
                "query": querystr,
                "Content-Type": "application/json",
            },
        )
        result = json.loads(r.text)

        received_date = []
        circular_nr = []

        for entry in result["data"]["allCirculars"]["edges"]:
            print(entry)
            """
            do some filtering based on subjects
            (there are erroneous event associations on the server)
            """
            if (
                "neutrino" in (subj := entry["node"]["subject"])
                and "high-energy" in subj
            ):
                received_date.append(entry["node"]["received"])
                circular_nr.append(entry["node"]["cid"])
        """
        I don't trust this webserver, let's go with the
        earliest GCN, not the last in the list
        """
        gcn_no = circular_nr[
            np.argmin([Time(i, format="isot").mjd for i in received_date])
        ]
        logging.info(f"GCN found ({gcn_no})")

        return gcn_no

    else:
        logging.warning(f"No GCN found for {base_nu_name}")

        return None


def get_latest_gcn():
    """
    Get the last circular
    """
    endpoint = (
        "https://heasarc.gsfc.nasa.gov/wsgi-scripts/tach/gcn_v2/tach.wsgi/graphql_fast"
    )
    querystr = (
        '{ allCirculars ( first:50after:"" ) { totalCount pageInfo{ '
        "hasNextPage hasPreviousPage startCursor endCursor } "
        "edges { node { id id_ received subject evtidCircular{ event } cid "
        "evtid oidCircular{ telescope detector "
        "oidEvent{ wavelength messenger } } } } } }"
    )

    r = requests.post(
        endpoint,
        data={
            "query": querystr,
            "Content-Type": "application/json",
        },
    )
    result = json.loads(r.text)

    received_date = []
    circular_nr = []

    for entry in result["data"]["allCirculars"]["edges"]:
        received_date.append(entry["node"]["received"])
        circular_nr.append(entry["node"]["cid"])

    latest_gcn_no = circular_nr[
        np.argmin([Time(i, format="isot").mjd for i in received_date])
    ]

    logging.info(f"Most recent GCN is {latest_gcn_no}")

    return latest_gcn_no


def parse_radec(string: str):
    """
    Find the RA and Dec in a string

    :param string: ra/dec string
    :return:
    """
    regex_findall = re.findall(r"[-+]?\d*\.\d+|\d+", string)

    if len(regex_findall) == 2:
        pos = float(regex_findall[0])
        pos_upper = None
        pos_lower = None
    elif len(regex_findall) == 4:
        pos = float(regex_findall[0])
        pos_upper = float(regex_findall[1])
        pos_lower = float(regex_findall[1])
    elif len(regex_findall) == 5:
        pos, pos_upper, pos_lower = regex_findall[0:3]
        pos = float(pos)
        pos_upper = float(pos_upper.replace("+", ""))
        pos_lower = float(pos_lower.replace("-", ""))
    else:
        raise ParsingError(f"Could not parse GCN ra and dec")

    return pos, pos_upper, pos_lower


def parse_gcn_circular(gcn_number: int):
    """
    Parses the handwritten text of a given GCN;
    extracts author, time and RA/Dec (with errors)
    """

    returndict = {}
    mainbody_starts_here = 999

    endpoint = f"https://gcn.nasa.gov/circulars/{gcn_number}.json"
    res = requests.get(endpoint)

    res_json = res.json()

    subject = res_json.get("subject")
    submitter = res_json.get("submitter")
    body = res_json.get("body")

    base = submitter.split("at")[0].split(" ")
    author = [x for x in base if x != ""][1]
    returndict.update({"author": author})

    name = subject.split(" - ")[0]
    returndict.update({"name": name})

    splittext = body.splitlines()
    splittext = list(filter(None, splittext))

    for i, line in enumerate(splittext):
        if (
            ("RA" in line or "Ra" in line)
            and ("DEC" in splittext[i + 1] or "Dec" in splittext[i + 1])
            and i < mainbody_starts_here
        ):
            ra, ra_upper, ra_lower = parse_radec(line)
            dec, dec_upper, dec_lower = parse_radec(splittext[i + 1])
            if ra_upper and ra_lower:
                ra_err = [ra_upper, -ra_lower]
            else:
                ra_err = [None, None]

            if dec_upper and dec_lower:
                dec_err = [dec_upper, -dec_lower]
            else:
                dec_err = [None, None]
            returndict.update(
                {"ra": ra, "ra_err": ra_err, "dec": dec, "dec_err": dec_err}
            )
            mainbody_starts_here = i + 2
        elif ("Time" in line or "TIME" in line) and i < mainbody_starts_here:
            raw_time = [
                x for x in line.split(" ") if x not in ["Time", "", "UT", "UTC"]
            ][1]
            raw_time = "".join(
                [x for x in raw_time if np.logical_or(x.isdigit(), x in [":", "."])]
            )
            raw_date = name.split("-")[1][:6]
            ut_time = f"20{raw_date[0:2]}-{raw_date[2:4]}-{raw_date[4:6]}T{raw_time}"
            time = Time(ut_time, format="isot", scale="utc")
            returndict.update({"time": time})

    return returndict



#!/usr/bin/env python3
# coding: utf-8

import logging

import healpy as hp
import numpy as np
import yaml
from astropy.time import Time
from tqdm import tqdm

from nuztf.base_scanner import BaseScanner
from nuztf.parse_nu_gcn import find_gcn_no, get_latest_gcn, parse_gcn_circular
from nuztf.paths import CONFIG_DIR


class NeutrinoScanner(BaseScanner):
    default_fritz_group = 221

    def __init__(
        self,
        nu_name: str = None,
        manual_args=None,
        gcn_no: int = None,
        cone_nside: int = 128,
        t_precursor: float = None,
        config: dict = None,
    ):
        self.logger = logging.getLogger(__name__)

        if config:
            self.config = config
        else:
            config_path = CONFIG_DIR.joinpath("nu_run_config.yaml")
            with open(config_path) as f:
                self.config = yaml.safe_load(f)

        self.prob_threshold = 0.9

        if manual_args is None:
            if nu_name is not None:
                gcn_no = find_gcn_no(nu_name)

            elif gcn_no is None:
                gcn_no = get_latest_gcn()

            gcn_info = parse_gcn_circular(gcn_no)

            self.logger.info(gcn_info)

            nu_name_gcn = gcn_info.get("name")
            if nu_name_gcn is None:
                raise ValueError(f"No GCN notice or circular found for {nu_name}.")
            else:
                nu_name = nu_name_gcn
            author = gcn_info["author"]
            ra = [gcn_info["ra"], gcn_info["ra_err"][0], gcn_info["ra_err"][1]]
            dec = [gcn_info["dec"], gcn_info["dec_err"][0], gcn_info["dec_err"][1]]
            nu_time = gcn_info["time"]

            if t_precursor is not None:
                nu_time = Time(nu_time.mjd - t_precursor, format="mjd")

        else:
            (nu_name, ra, dec, nu_time) = manual_args
            author = None

        self.nu_name = nu_name
        self.author = author
        self.gcn_no = gcn_no
        self.dist = None

        self.logger.info(f"Neutrino time: {nu_time}")

        self.ra_max = float(max(ra[1:]) + ra[0])
        self.ra_min = float(min(ra[1:]) + ra[0])
        self.dec_max = float(max(dec[1:]) + dec[0])
        self.dec_min = float(min(dec[1:]) + dec[0])

        self.logger.info(f"Coordinates: RA = {ra[0]} ({self.ra_min} - {self.ra_max})")
        self.logger.info(
            f"Coordinates: Dec = {dec[0]} ({self.dec_min} - {self.dec_max})"
        )

        BaseScanner.__init__(
            self,
            t_min=nu_time,
            run_config=self.config,
            cone_nside=cone_nside,
        )
        self.prob_threshold = 0.9
        self.rectangular_area = (
            (self.ra_max - self.ra_min)
            * (self.dec_max - self.dec_min)
            * abs(np.cos(np.radians(dec[0])))
        )

        self.logger.info(f"Projected Area: {self.rectangular_area:.3f} sq. deg.")

    def get_name(self):
        """ """
        return self.nu_name

    def get_full_name(self):
        """ """
        return (
            f"neutrino event {self.get_name()} "
            f"({self.author} et. al, GCN {self.gcn_no})"
        )

    @staticmethod
    def get_obs_line():
        """ """
        return "Each exposure was 300s with a typical depth of 21.0 mag."

    @staticmethod
    def remove_variability_line():
        """ """
        return ""

    def filter_f_no_prv(self, res: dict):
        """ """

        ztf_id = res["objectId"]

        # Positive detection
        if res["candidate"]["isdiffpos"] not in ["t", "1"]:
            self.logger.debug(f"❌ {ztf_id}: Negative subtraction")
            return False

        try:
            if res["candidate"]["drb"] < 0.3:
                self.logger.debug(f"❌ {ztf_id}: DRB too low")
                return False
        except (KeyError, TypeError) as e:
            pass

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.debug(f"❌ {ztf_id}: Not in contour")
            return False

        endhist = res["candidate"]["jdendhist"]
        starthist = res["candidate"]["jdstarthist"]

        if endhist == starthist:
            self.logger.debug(f"❌ {ztf_id}: One detection only")
            return False

        # Require 2 detections separated by 15 mins
        if (endhist - starthist) < 0.01:
            self.logger.debug(
                f"❌ {ztf_id}: Does have 2 detections, but these are not separated by >15 mins (delta t = {(endhist-starthist)*1440:.0f} min)"
            )
            return False

        self.logger.debug(f"✅ {ztf_id}: Passes first filtering stage (no prv).")
        return True

    def filter_f_history(self, res: dict):
        """Filter based on 2 detection requirement and probability contour requirement"""

        ztf_id = res["objectId"]

        n_detections = len(
            [x for x in res["prv_candidates"] if "isdiffpos" in x.keys()]
        )

        if n_detections < 1:
            self.logger.info(f"❌ {ztf_id}: Has insufficient detections")
            return False

        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            self.logger.info(f"❌ {ztf_id}: Not in contour")
            return False

        self.logger.debug(f"✅ {ztf_id}: Passed the history filtering stage")

        return True

    def find_cone_coords(self):
        """ """
        cone_coords = []
        cone_ids = []

        scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))

        self.logger.info("Finding search pixels:")

        for i in tqdm(range(hp.nside2npix(self.cone_nside))):
            ra, dec = self.extract_ra_dec(self.cone_nside, i)
            ra_rad = np.radians(ra)
            dec_rad = np.radians(dec)
            if np.logical_and(
                ra > self.ra_min - scan_radius, ra < self.ra_max + scan_radius
            ):
                if np.logical_and(
                    dec > self.dec_min - scan_radius,
                    dec < self.dec_max + scan_radius,
                ):
                    cone_coords.append((ra_rad, dec_rad))
                    cone_ids.append(i)

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", float), ("dec", float)])
        )

        return cone_ids, cone_coords

    def in_contour(self, ra_deg, dec_deg):
        in_ra = np.logical_and(ra_deg > self.ra_min, ra_deg < self.ra_max)
        in_dec = np.logical_and(dec_deg > self.dec_min, dec_deg < self.dec_max)

        return np.logical_and(in_ra, in_dec)

    def unpack_skymap(self, skymap=None, output_nside: None | int = None):
        """ """
        output_nside = 1024

        map_coords = []
        pixel_nos = []

        center_ra = np.radians(np.mean([self.ra_max, self.ra_min]))
        center_dec = np.radians(np.mean([self.dec_max, self.dec_min]))
        rad = (
            np.radians(max(self.ra_max - self.ra_min, self.dec_max - self.dec_min))
            / 2.0
        )

        nearish_pixels = list(
            hp.query_disc(
                nside=output_nside,
                vec=hp.ang2vec(np.pi / 2.0 - center_dec, center_ra),
                radius=rad,
                nest=True,
            )
        )

        for i in tqdm(nearish_pixels):
            ra, dec = self.extract_ra_dec(output_nside, i)
            if self.in_contour(ra, dec):
                map_coords.append((ra, dec))
                pixel_nos.append(i)

        map_probs = np.ones_like(pixel_nos, dtype=float)
        map_probs /= np.sum(map_probs)

        key = "PROB"

        data = np.zeros(hp.nside2npix(output_nside), dtype=np.dtype([(key, float)]))
        data[np.array(pixel_nos)] = map_probs

        total_pixel_area = hp.nside2pixarea(output_nside, degrees=True) * float(
            len(map_coords)
        )

        return (
            map_coords,
            pixel_nos,
            output_nside,
            map_probs,
            data,
            total_pixel_area,
            key,
        )



import logging
import os
import pickle
from pathlib import Path

from gwemopt.ztf_tiling import get_quadrant_ipix
from tqdm import tqdm
from ztfquery.fields import FIELD_DATAFRAME

from nuztf.paths import FLATPIX_CACHE_DIR


def get_flatpix_path(nside: int) -> Path:
    """
    Flatpix lookup table for nside

    :param nside: nside of healpix
    :return: path to flatpix lookup table
    """
    return FLATPIX_CACHE_DIR.joinpath(f"ztf_fields_ipix_nside={nside}.pickle")


def get_nested_pix_path(nside: int) -> Path:
    """
    Nested pix lookup table for nside

    :param nside: nside of healpix
    :return: path to nested pix lookup table
    """

    outfile = FLATPIX_CACHE_DIR.joinpath(f"ztf_fields_nested_ipix_nside={nside}.pickle")
    return outfile


def generate_flatpix_file(nside: int, logger=logging.getLogger(__name__)):
    """
    Generate and save the fields-healpix lookup table

    :param nside: nside of healpix
    :param logger: logger
    :return: None
    """

    logger.info(f"Generating field-healpix lookup table for nside={nside}")

    field_dataframe = FIELD_DATAFRAME.reset_index()

    fields = field_dataframe["ID"].values
    ras = field_dataframe["RA"].values
    decs = field_dataframe["Dec"].values

    flat_pix_dict = dict()
    nested_pix_dict = dict()

    for i, field in tqdm(enumerate(fields), total=len(fields)):
        ra = ras[i]
        dec = decs[i]
        pix = get_quadrant_ipix(nside, ra, dec)

        flat_pix = []

        for sub_list in pix:
            for p in sub_list:
                flat_pix.append(p)

        flat_pix = list(set(flat_pix))
        flat_pix_dict[field] = flat_pix
        nested_pix_dict[field] = pix

    outfile = get_flatpix_path(nside=nside)

    logger.info(f"Saving to {outfile}")

    with open(outfile, "wb") as f:
        pickle.dump(flat_pix_dict, f)

    outfile = get_nested_pix_path(nside=nside)

    logger.info(f"Saving to {outfile}")

    with open(outfile, "wb") as f:
        pickle.dump(nested_pix_dict, f)


def get_flatpix(nside: int, logger=logging.getLogger(__name__)):
    """
    Get the fields-healpix lookup table

    :param nside: nside of healpix
    :param logger: logger
    :return: flatpix lookup table
    """
    infile = get_flatpix_path(nside=nside)

    # Generate a lookup table for field healpix
    # if none exists (because this is computationally costly)
    if not infile.exists():
        generate_flatpix_file(nside=nside, logger=logger)

    logger.info(f"Loading from {infile}")

    with open(infile, "rb") as f:
        field_pix = pickle.load(f)

    return field_pix


def get_nested_pix(nside: int, logger=logging.getLogger(__name__)):
    """
    Nested Flatpix lookup table for nside

    :param nside: nside of healpix
    :param logger: logger
    :return: flatpix lookup table
    """
    infile = get_nested_pix_path(nside=nside)

    # Generate a lookup table for field healpix
    # if none exists (because this is computationally costly)
    if not os.path.isfile(infile):
        generate_flatpix_file(nside=nside, logger=logger)

    logger.info(f"Loading from {infile}")

    with open(infile, "rb") as f:
        field_pix = pickle.load(f)

    return field_pix



import os
from pathlib import Path

CONFIG_DIR = Path(__file__).parent.joinpath("config")

NUZTF_OUTPUT_DIR_STR = os.environ.get("NUZTF_DIR")

if NUZTF_OUTPUT_DIR_STR is None:
    NUZTF_OUTPUT_DIR = Path.home() / "Data" / "nuztf"
else:
    NUZTF_OUTPUT_DIR = Path(NUZTF_OUTPUT_DIR_STR)

SKYMAP_DIR = NUZTF_OUTPUT_DIR / "skymaps"
SKYMAP_DIR.mkdir(exist_ok=True, parents=True)

RESULTS_DIR = NUZTF_OUTPUT_DIR / "results"
CACHE_DIR = NUZTF_OUTPUT_DIR / "cache"
BASE_CANDIDATE_DIR = CACHE_DIR / "candidates"

CROSSMATCH_CACHE = CACHE_DIR / "crossmatch"
CROSSMATCH_CACHE.mkdir(exist_ok=True, parents=True)

CUTOUT_CACHE_DIR = CACHE_DIR / "cutouts"
CUTOUT_CACHE_DIR.mkdir(exist_ok=True, parents=True)

FLATPIX_CACHE_DIR = CACHE_DIR / "flatpix"
FLATPIX_CACHE_DIR.mkdir(exist_ok=True, parents=True)

PREPROCESSED_CACHE_DIR = CACHE_DIR / "preprocessed"
PREPROCESSED_CACHE_DIR.mkdir(exist_ok=True, parents=True)



#!/usr/bin/env python
# coding: utf-8

import logging
import os
import warnings

from ztfquery import io

# Manage ztfquery logins from environment variables


def load_credentials(name: str, token_based: bool = False):
    """ZTFquery wrapper for loading credentials."""
    return io._load_id_(name, token_based=token_based)


try:
    io.set_account(
        "irsa", username=os.environ["IRSA_USER"], password=os.environ["IRSA_PASSWORD"]
    )
    logging.info('Set up "irsa" credentials')

except KeyError:
    logging.info(
        'No Credentials for "irsa" found in environment' "Assuming they are set."
    )

try:
    io.set_account(
        "skyvision",
        username=os.environ["SKYVISION_USER"],
        password=os.environ["SKYVISION_PASSWORD"],
    )
    logging.info('Set up "skyvision" credentials')

except KeyError:
    logging.info(
        'No Credentials for "skyvision" found in environment' "Assuming they are set."
    )

try:
    io.set_account(
        "ipacdepot",
        username=os.environ["DEPOT_USER"],
        password=os.environ["DEPOT_PASSWORD"],
    )
    logging.info('Set up "DEPOT" credentials')

except KeyError:
    logging.info(
        'No Credentials for "DEPOT" found in environment' "Assuming they are set."
    )

try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        io.set_account(
            "ampel_api_archive_token",
            username=os.environ["AMPEL_API_ARCHIVE_TOKEN_USER"],
            password=os.environ["AMPEL_API_ARCHIVE_TOKEN_PASSWORD"],
        )
        logging.info('Set up "ampel_api_archive_token" credentials')

except KeyError:
    logging.info("No Token for AMPEL API found in environment" "Assume they are set.")

try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        io.set_account(
            "tns_api_token", token=os.environ["TNS_API_TOKEN"], token_based=True
        )
        logging.info('Set up "tns_api_token" credentials')

except KeyError:
    logging.info("No Token for TNS API found in environment" "Assume it is set.")

try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        io.set_account(
            "desy_cloud_token", token=os.environ["DESY_CLOUD_TOKEN"], token_based=True
        )
        logging.info('Set up "desy_cloud_token" credentials')

except KeyError:
    logging.info("No Token for DESY Cloud found in environment" "Assume it is set.")

try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        io.set_account("fritz", token=os.environ["FRITZ_TOKEN"], token_based=True)
        logging.info('Set up "fritz" credentials')

except KeyError:
    logging.info("No Token for Fritz API found in environment" "Assume it is set.")
